/*
This is free and unencumbered software released into the public domain.

Anyone is free to copy, modify, publish, use, compile, sell, or
distribute this software, either in source code form or as a compiled
binary, for any purpose, commercial or non-commercial, and by any
means.

In jurisdictions that recognize copyright laws, the author or authors
of this software dedicate any and all copyright interest in the
software to the public domain. We make this dedication for the benefit
of the public at large and to the detriment of our heirs and
successors. We intend this dedication to be an overt act of
relinquishment in perpetuity of all present and future rights to this
software under copyright law.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

For more information, please refer to <http://unlicense.org/>
*/

// Written by Richard Mitton
// http://www.codersnotes.com/notes/warnock-subdivision-for-deferred-lighting

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <algorithm>
#include <SDL/SDL.h>

#ifdef EMSCRIPTEN
#include <emscripten.h>
#endif

#define MAX_RECTANGLES 26

struct Rectangle
{
	// co-ordinates (includes x0, excludes x1)
	int x0, y0, x1, y1;
	
	// payload data attached to our rectangle:
	uint32_t color;
	int id;
};

// Helpers ------------------------------------------------------------------

// Globals for our demo.
bool dragging = false;
SDL_Surface *screen;
int mx, my, mb;
int numRectangles = 0;
Rectangle rectangles[MAX_RECTANGLES];

unsigned char fontdata[12*256] = {
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,// ' '
  0, 12, 30, 30, 30, 12, 12,  0, 12, 12,  0,  0,// '!'
  0,102,102,102, 36,  0,  0,  0,  0,  0,  0,  0,// '"'
  0, 54, 54,127, 54, 54, 54,127, 54, 54,  0,  0,// '#'
 12, 12, 62,  3,  3, 30, 48, 48, 31, 12, 12,  0,// '$'
  0,  0,  0, 35, 51, 24, 12,  6, 51, 49,  0,  0,// '%'
  0, 14, 27, 27, 14, 95,123, 51, 59,110,  0,  0,// '&'
  0, 12, 12, 12,  6,  0,  0,  0,  0,  0,  0,  0,// '''
  0, 48, 24, 12,  6,  6,  6, 12, 24, 48,  0,  0,// '('
  0,  6, 12, 24, 48, 48, 48, 24, 12,  6,  0,  0,// ')'
  0,  0,  0,102, 60,255, 60,102,  0,  0,  0,  0,// '*'
  0,  0,  0, 24, 24,126, 24, 24,  0,  0,  0,  0,// '+'
  0,  0,  0,  0,  0,  0,  0,  0, 28, 28,  6,  0,// ','
  0,  0,  0,  0,  0,127,  0,  0,  0,  0,  0,  0,// '-'
  0,  0,  0,  0,  0,  0,  0,  0, 28, 28,  0,  0,// '.'
  0,  0, 64, 96, 48, 24, 12,  6,  3,  1,  0,  0,// '/'
  0, 62, 99,115,123,107,111,103, 99, 62,  0,  0,// '0'
  0,  8, 12, 15, 12, 12, 12, 12, 12, 63,  0,  0,// '1'
  0, 30, 51, 51, 48, 24, 12,  6, 51, 63,  0,  0,// '2'
  0, 30, 51, 48, 48, 28, 48, 48, 51, 30,  0,  0,// '3'
  0, 48, 56, 60, 54, 51,127, 48, 48,120,  0,  0,// '4'
  0, 63,  3,  3,  3, 31, 48, 48, 51, 30,  0,  0,// '5'
  0, 28,  6,  3,  3, 31, 51, 51, 51, 30,  0,  0,// '6'
  0,127, 99, 99, 96, 48, 24, 12, 12, 12,  0,  0,// '7'
  0, 30, 51, 51, 55, 30, 59, 51, 51, 30,  0,  0,// '8'
  0, 30, 51, 51, 51, 62, 24, 24, 12, 14,  0,  0,// '9'
  0,  0,  0, 28, 28,  0,  0, 28, 28,  0,  0,  0,// ':'
  0,  0,  0, 28, 28,  0,  0, 28, 28, 24, 12,  0,// ';'
  0, 48, 24, 12,  6,  3,  6, 12, 24, 48,  0,  0,// '<'
  0,  0,  0,  0,126,  0,126,  0,  0,  0,  0,  0,// '='
  0,  6, 12, 24, 48, 96, 48, 24, 12,  6,  0,  0,// '>'
  0, 30, 51, 48, 24, 12, 12,  0, 12, 12,  0,  0,// '?'
  0, 62, 99, 99,123,123,123,  3,  3, 62,  0,  0,// '@'
  0, 12, 30, 51, 51, 51, 63, 51, 51, 51,  0,  0,// 'A'
  0, 63,102,102,102, 62,102,102,102, 63,  0,  0,// 'B'
  0, 60,102, 99,  3,  3,  3, 99,102, 60,  0,  0,// 'C'
  0, 31, 54,102,102,102,102,102, 54, 31,  0,  0,// 'D'
  0,127, 70,  6, 38, 62, 38,  6, 70,127,  0,  0,// 'E'
  0,127,102, 70, 38, 62, 38,  6,  6, 15,  0,  0,// 'F'
  0, 60,102, 99,  3,  3,115, 99,102,124,  0,  0,// 'G'
  0, 51, 51, 51, 51, 63, 51, 51, 51, 51,  0,  0,// 'H'
  0, 30, 12, 12, 12, 12, 12, 12, 12, 30,  0,  0,// 'I'
  0,120, 48, 48, 48, 48, 51, 51, 51, 30,  0,  0,// 'J'
  0,103,102, 54, 54, 30, 54, 54,102,103,  0,  0,// 'K'
  0, 15,  6,  6,  6,  6, 70,102,102,127,  0,  0,// 'L'
  0, 99,119,127,127,107, 99, 99, 99, 99,  0,  0,// 'M'
  0, 99, 99,103,111,127,123,115, 99, 99,  0,  0,// 'N'
  0, 28, 54, 99, 99, 99, 99, 99, 54, 28,  0,  0,// 'O'
  0, 63,102,102,102, 62,  6,  6,  6, 15,  0,  0,// 'P'
  0, 28, 54, 99, 99, 99,115,123, 62, 48,120,  0,// 'Q'
  0, 63,102,102,102, 62, 54,102,102,103,  0,  0,// 'R'
  0, 30, 51, 51,  3, 14, 24, 51, 51, 30,  0,  0,// 'S'
  0, 63, 45, 12, 12, 12, 12, 12, 12, 30,  0,  0,// 'T'
  0, 51, 51, 51, 51, 51, 51, 51, 51, 30,  0,  0,// 'U'
  0, 51, 51, 51, 51, 51, 51, 51, 30, 12,  0,  0,// 'V'
  0, 99, 99, 99, 99,107,107, 54, 54, 54,  0,  0,// 'W'
  0, 51, 51, 51, 30, 12, 30, 51, 51, 51,  0,  0,// 'X'
  0, 51, 51, 51, 51, 30, 12, 12, 12, 30,  0,  0,// 'Y'
  0,127,115, 25, 24, 12,  6, 70, 99,127,  0,  0,// 'Z'
  0, 60, 12, 12, 12, 12, 12, 12, 12, 60,  0,  0,// '['
  0,  0,  1,  3,  6, 12, 24, 48, 96, 64,  0,  0,// '\'
  0, 60, 48, 48, 48, 48, 48, 48, 48, 60,  0,  0,// ']'
  8, 28, 54, 99,  0,  0,  0,  0,  0,  0,  0,  0,// '^'
  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,255,  0,// '_'
 12, 12, 24,  0,  0,  0,  0,  0,  0,  0,  0,  0,// '`'
  0,  0,  0,  0, 30, 48, 62, 51, 51,110,  0,  0,// 'a'
  0,  7,  6,  6, 62,102,102,102,102, 59,  0,  0,// 'b'
  0,  0,  0,  0, 30, 51,  3,  3, 51, 30,  0,  0,// 'c'
  0, 56, 48, 48, 62, 51, 51, 51, 51,110,  0,  0,// 'd'
  0,  0,  0,  0, 30, 51, 63,  3, 51, 30,  0,  0,// 'e'
  0, 28, 54,  6,  6, 31,  6,  6,  6, 15,  0,  0,// 'f'
  0,  0,  0,  0,110, 51, 51, 51, 62, 48, 51, 30,// 'g'
  0,  7,  6,  6, 54,110,102,102,102,103,  0,  0,// 'h'
  0, 24, 24,  0, 30, 24, 24, 24, 24,126,  0,  0,// 'i'
  0, 48, 48,  0, 60, 48, 48, 48, 48, 51, 51, 30,// 'j'
  0,  7,  6,  6,102, 54, 30, 54,102,103,  0,  0,// 'k'
  0, 30, 24, 24, 24, 24, 24, 24, 24,126,  0,  0,// 'l'
  0,  0,  0,  0, 63,107,107,107,107, 99,  0,  0,// 'm'
  0,  0,  0,  0, 31, 51, 51, 51, 51, 51,  0,  0,// 'n'
  0,  0,  0,  0, 30, 51, 51, 51, 51, 30,  0,  0,// 'o'
  0,  0,  0,  0, 59,102,102,102,102, 62,  6, 15,// 'p'
  0,  0,  0,  0,110, 51, 51, 51, 51, 62, 48,120,// 'q'
  0,  0,  0,  0, 55,118,110,  6,  6, 15,  0,  0,// 'r'
  0,  0,  0,  0, 30, 51,  6, 24, 51, 30,  0,  0,// 's'
  0,  0,  4,  6, 63,  6,  6,  6, 54, 28,  0,  0,// 't'
  0,  0,  0,  0, 51, 51, 51, 51, 51,110,  0,  0,// 'u'
  0,  0,  0,  0, 51, 51, 51, 51, 30, 12,  0,  0,// 'v'
  0,  0,  0,  0, 99, 99,107,107, 54, 54,  0,  0,// 'w'
  0,  0,  0,  0, 99, 54, 28, 28, 54, 99,  0,  0,// 'x'
  0,  0,  0,  0,102,102,102,102, 60, 48, 24, 15,// 'y'
  0,  0,  0,  0, 63, 49, 24,  6, 35, 63,  0,  0,// 'z'
  0, 56, 12, 12,  6,  3,  6, 12, 12, 56,  0,  0,// '{'
  0, 24, 24, 24, 24,  0, 24, 24, 24, 24,  0,  0,// '|'
  0,  7, 12, 12, 24, 48, 24, 12, 12,  7,  0,  0,// '}'
  0,206, 91,115,  0,  0,  0,  0,  0,  0,  0,  0,// '~'
};

Rectangle fixRect(Rectangle r)
{
	// Make sure they're the right way around:
	if (r.x1 < r.x0)
		std::swap(r.x1, r.x0);
	if (r.y1 < r.y0)
		std::swap(r.y1, r.y0);
	return r;
}

// Prints a string on-screen at the given co-ordinates.
void print(int dx, int dy, uint32_t col, const char *str)
{
	uint32_t *pix = (uint32_t *)screen->pixels;
	uint32_t rgb = SDL_MapRGB(screen->format, (col>>16)&0xff, (col>>8)&0xff, (col>>0)&0xff);
	size_t pitch = screen->pitch / 4;
	while (*str)
	{
		char c = *str++;
		for (int y=0;y<12;y++)
		{
			for (int x=0;x<8;x++)
			{
				int px = x + dx, py = y + dy;
				if (px >= 0 && py >= 0 && px < screen->w && py < screen->h)
				{
					if (fontdata[(c - 32) * 12 + y] & (1 << x))
						pix[py*pitch + px] = rgb;
				}
			}
		}
		dx += 8;
	}
}

// Draws a rectangle outline (exclusive).
void rect(int x0, int y0, int x1, int y1, uint32_t col)
{
	uint32_t *pix = (uint32_t *)screen->pixels;
	uint32_t rgb = SDL_MapRGB(screen->format, (col>>16)&0xff, (col>>8)&0xff, (col>>0)&0xff);
	size_t pitch = screen->pitch / 4;
	for (int x=x0;x<x1;x++)
	{
		pix[y0*pitch+x] = rgb;
		pix[y1*pitch+x] = rgb;
	}
	for (int y=y0;y<y1;y++)
	{
		pix[y*pitch+x0] = rgb;
		pix[y*pitch+x1] = rgb;
	}
}

// Draws a filled rectangle (exclusive).
void rectfill(int x0, int y0, int x1, int y1, uint32_t col)
{
	uint32_t *pix = (uint32_t *)screen->pixels;
	uint32_t rgb = SDL_MapRGB(screen->format, (col>>16)&0xff, (col>>8)&0xff, (col>>0)&0xff);
	size_t pitch = screen->pitch / 4;
	for (int y=y0;y<y1;y++)
		for (int x=x0;x<x1;x++)
			pix[y*pitch+x] = rgb;
}

// Algorithm ----------------------------------------------------------------

// The state we track for each window.
struct State
{
	Rectangle *list;
	int count;
	int x0, y0, x1, y1;
};

// Given a window and a list of rectangles that fully cover
// that window, draw them all.
// In a deferred renderer, we would here issue draw calls
// to do the actual lighting for those lights.
// For this demo, we'll just draw the window outline and
// print some text info.
void drawRectangleList(State &s)
{
	// Draw the window edge.	
	rect(s.x0, s.y0, s.x1, s.y1, 0xc0c0c0);
	
	// If the cursor is over it, highlight it with more information.
	if (mx >= s.x0 && my >= s.y0 && mx < s.x1 && my < s.y1)
	{
		// Fill in the window so we can see it.
		rectfill(s.x0, s.y0, s.x1, s.y1, 0xf0f0f0);

		// Print a list of all the rectangles.
		char tmp[256];
		
		if (s.count)
		{
			strcpy(tmp, "[ ");
			for (int n=0;n<s.count;n++)
			{
				if (n) strcat(tmp, ",");
				sprintf(tmp+strlen(tmp), "%c", s.list[n].id);
			}
			strcat(tmp, " ]");
		} else {
			strcpy(tmp, "<empty>");
		}
		print(mx+30, my+30, 0x808080, tmp);
	}
}

// Renders a set of overlapping rectangles on-screen by
// recursively subdividing them into regions where
// each region contains no rectangle edges.
void warnockSubdivide(Rectangle *list, int count)
{
	// Initialize the stack to the whole window, and all the rectangles.
	int stackSize = 1;
	State stack[MAX_RECTANGLES*4];
	stack[0] = (State) { list, count, 0, 0, screen->w, screen->h };
	
	while (stackSize > 0)
	{
		// Pop a new window off the stack.
		State s = stack[--stackSize];
		
		// Check if our current window is OK to render to:
		if (s.x1 <= s.x0 || s.y1 <= s.y0)
			continue;
		
		// If we need to recurse, we'll need to decide on a point
		// within the window to split at. We really want it to be
		// on the edge of a rectangle. We'll fill it in blank for
		// now and get a better guess as we test each rectangle.
		int splitX = s.x0, splitY = s.y0;

		// Let's treat our list of rectangles as a "triage table".
		// This means we consider the list as 3 categories:
		//
		// |---outside---|---partial---|---remaining---|---inside---|
		//                             ^- read cursor
		//
		// The 'remaining' list initially contains every item, with
		// the other lists being empty.
		// We process through the 'remaining' list, and move each item
		// into it's appropriate category. We can do this entirely
		// in-place within this one array.
		// Once the 'remaining' list is empty (i.e. the read cursor
		// is equal to the start of the 'inside' list) then we're done.
		int endOutside = 0;        // index of the end of the 'outside' list.
		int startInside = s.count; // index of the start of the 'inside' list.

		// Classify each rectangle into one of three categories:
		// a) The rectangle is entirely outside our window
		// b) The rectangle entirely covers our window
		// c) The rectangle partially intersects our window
		for (int n=0;n<startInside;)
		{
			const Rectangle &r = s.list[n];
			
			// Compute the in/out clipping codes.
			// A code has one of four bits set depending
			// on whether that edge intersects our window.
			int outcode = 0, incode = 0;
			if (r.x0 >= s.x1) outcode |= 1;
			if (r.y0 >= s.y1) outcode |= 2;
			if (r.x1 <= s.x0) outcode |= 4;
			if (r.y1 <= s.y0) outcode |= 8;
			if (r.x0 <= s.x0) incode |= 1;
			if (r.y0 <= s.y0) incode |= 2;
			if (r.x1 >= s.x1) incode |= 4;
			if (r.y1 >= s.y1) incode |= 8;

			if (outcode != 0) {
				// Rectangle is entirely outside this window.
				// Move it to the 'outside' list.
				std::swap(s.list[n++], s.list[endOutside++]);
			} else if (incode == 0xf) {
				// Rectangle entirely covers this window.
				// Move it to the 'inside' list.
				std::swap(s.list[n], s.list[--startInside]);
			} else {
				// This rectangle partially covers, so we'll
				// see if one of it's edges can be used as a suitable
				// split point in case we need to recurse later.
				// Another option is to simply split the window exactly
				// in half - this is easier to do, but generates more
				// output regions than necessary.
				if (!( incode & 1))
					splitX = r.x0;
				if (!( incode & 2))
					splitY = r.y0;
				if (!( incode & 4))
					splitX = r.x1;
				if (!( incode & 8))
					splitY = r.y1;

				// Leave this rectangle where it is and
				// grow the 'partial' list to cover it.
				n++;
			}
		}
	
		// Throw away the ones that are entirely outside, giving
		// us a list of ones that are either entirely or partially
		// within our window.
		int numChild = s.count - endOutside;
		Rectangle *children = s.list + endOutside;

		// If we have any partials, we recurse with smaller windows.
		int numPartial = startInside - endOutside;
		if (numPartial > 0)
		{
			stack[stackSize++] = (State){ children, numChild, s.x0, s.y0, splitX, splitY };
			stack[stackSize++] = (State){ children, numChild, splitX, s.y0, s.x1, splitY };
			stack[stackSize++] = (State){ children, numChild, s.x0, splitY, splitX, s.y1 };
			stack[stackSize++] = (State){ children, numChild, splitX, splitY, s.x1, s.y1 };
		} else {
			// Otherwise we now know that everything in this list
			// entirely covers our window, so we can just stop here and draw them all.
			State final = { children, numChild, s.x0, s.y0, s.x1, s.y1 };
			drawRectangleList(final);
		}
	}
}

void render()
{   
	SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0xff, 0xff, 0xff));
	SDL_LockSurface(screen);

	// Add all our target rectangles into a temporary list:
	Rectangle list[MAX_RECTANGLES];
	for (int n=0;n<numRectangles;n++)
		list[n] = fixRect(rectangles[n]);
	
	// We'll use a Fisher-Yates shuffle to organize them into
	// a random order. This aids performance as it means
	// that when we pick the split point, we'll get a much better
	// chance of it being a nice even split, with the
	// same number of rectangles on each side.
	srand(1);
	for (int i=numRectangles-1;i>=1;i--)
	{
		int j = rand() % i;
		std::swap(list[i], list[j]);
	}
	
	// Do the actual algorithm.
	warnockSubdivide(list, numRectangles);

	print(10, 10, 0xc07000, "Click+drag to draw new rectangles.");
	print(10, 30, 0xc07000, "Space to reset.");
	
	// Draw the original rectangles on top, so we can 
	// visualize them for this demo.
	for (int n=0;n<numRectangles;n++)
	{
		char tmp[16];
		Rectangle r = fixRect(rectangles[n]);
		sprintf(tmp, "%c", r.id);
		rect(r.x0, r.y0, r.x1, r.y1, r.color);
		if (!dragging || n != numRectangles - 1)
			print(r.x0 + 5, r.y0 + 3, r.color, tmp);
	}
	
	rect(0, 0, screen->w-1, screen->h-1, 0x000000);
	SDL_UnlockSurface(screen);
	SDL_Flip(screen); 
}

uint32_t hsv_to_rgb(float h, float s, float v)
{
	int h_i = (int)(h*6);
	float f = h*6 - h_i;
	float p = v * (1 - s);
	float q = v * (1 - f*s);
	float t = v * (1 - (1 - f) * s);
	int r, g, b;
	switch (h_i) {
	case 0: r = v*255; g = t*255; b = p*255; break;
	case 1: r = q*255; g = v*255; b = p*255; break;
	case 2: r = p*255; g = v*255; b = t*255; break;
	case 3: r = p*255; g = q*255; b = v*255; break;
	case 4: r = t*255; g = p*255; b = v*255; break;
	case 5: r = v*255; g = p*255; b = q*255; break;
	}
	return (r << 16) | (g << 8) | b;
}
	
uint32_t randomColor()
{
	static float hue = 0.5f;
	hue += 0.618033988749895;
	return hsv_to_rgb(fmod(hue, 1.0), 0.75, 0.95);
}

void mainloop()
{
	SDL_Event event;
	while (SDL_PollEvent(&event)) 
	{
		switch (event.type) 
		{
			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_SPACE)
					numRectangles = 0;
				break;
			case SDL_KEYUP:
				if (event.key.keysym.sym == SDLK_ESCAPE)
					exit(0);
				break;
			case SDL_QUIT:
				exit(0);
		}
	}

	mb = SDL_GetMouseState(&mx, &my) & SDL_BUTTON(SDL_BUTTON_LEFT);

	// Handle dragging new windows out.
	if (mb)
	{
		if (!dragging && numRectangles < MAX_RECTANGLES)
		{
			Rectangle *r = &rectangles[numRectangles++];
			r->x0 = r->x1 = mx;
			r->y0 = r->y1 = my;
			r->color = randomColor();
			r->id = 'A' + numRectangles-1;
			dragging = true;
		}
		if (dragging)
		{
			Rectangle *r = &rectangles[numRectangles-1];
			r->x1 = mx;
			r->y1 = my;
		}
	} else {
		dragging = false;
	}

	render();
}

int main(int argc, char *argv[])
{
	SDL_Init(SDL_INIT_VIDEO);
	atexit(SDL_Quit);
	screen = SDL_SetVideoMode(640, 480, 32, SDL_SWSURFACE);  
	
	// Add some default rectangles.
	numRectangles = 2;
	rectangles[0] = (Rectangle) { 50, 50, 100,100, 0xff0000, 'A' };
	rectangles[1] = (Rectangle) { 30,157, 180,180, 0x0040ff, 'B' };	

#ifdef EMSCRIPTEN
	emscripten_set_main_loop(mainloop, 30, 0);
#else
	while (1)
		mainloop();
#endif    

	return 0;
}
